{***************************************************************
 *
 * Unit Name: ADO_Func
 * Purpose  : Provide functions for ADO objects
 * Author   : Andrew Baylis
 * Date     : 11/07/2001
 * History  : 1) Provide functions to convert a ConnectionString to
                 a TStrings list and back again.
              2) Provide a function to access the indexfieldnames as a stringlist
 *
 ****************************************************************}

unit ADO_Func;

interface

uses Classes,sysutils,db,ADOInt,ADODB;

const
   Jet4 = 'Microsoft.Jet.OLEDB.4.0';
   Engine = 'Jet OLEDB:Engine Type';
   cAutoInc = 'AutoIncrement';
   cAutoIncrement = 'Increment';
   cAutoSeed = 'Seed';
   cNullable = 'Nullable';
   cDefault = 'Default';
   cJETAllowZeroLength = 'Jet OLEDB:Allow Zero Length';
   cJETValidateRule =    'Jet OLEDB:Column Validation Rule';
   cJETValidateText =    'Jet OLEDB:Column Validation Text';
   cJETHyperlink =       'Jet OLEDB:Hyperlink';
   cJETGUIDAuto =        'Jet OLEDB:AutoGenerate';
   cJETRemoteTable =     'Jet OLEDB:Remote Table Name';
   cJETLinkProvider =    'Jet OLEDB:Link Provider String';
   cJETLinkSource =      'Jet OLEDB:Link Datasource';
   cJETCreateLink =      'Jet OLEDB:Create Link';
   cJETEncrypt =         'Jet OLEDB:Encrypt Database';

   sNullable = [ftString, ftVarBytes, ftBlob, ftMemo, ftGraphic, ftFmtMemo,
   ftParadoxOle, ftDBaseOle, ftTypedBinary, ftFixedChar, ftWideString, ftADT,
   ftArray, ftReference, ftDataSet, ftOraBlob, ftOraClob, ftVariant];

   {ftBoolean for some reason doesn't like Default properties set}
   sValidate = [ftString, ftSmallint, ftInteger, ftWord, {ftBoolean, }ftFloat,
   ftCurrency, ftBCD, ftDate, ftTime, ftDateTime,  ftWideString, ftLargeint];

function ConvertListToString(const List: TStrings; delim: Char = ';';finalDelim: Boolean = false): string;
procedure ConvertStringToList(const s: string; List: TStrings; delim:Char = ';');
function ADOTypeToFieldType(const ADOType: DataTypeEnum; EnableBCD: Boolean = True): TFieldType;
function FieldTypeToADOType(const FieldType: TFieldType): DataTypeEnum;

procedure IndexFieldsToStrings(ADOTab:TCustomADODataSet;List:TStringList);

procedure StringsToIndexFields(ADOTab:TCustomADODataSet;List:TStringlist);

implementation

function ConvertListToString(const List: TStrings; delim: Char = ';';finalDelim: Boolean = false): string;
var
  s: string;
  i: Integer;
begin
  s := '';
  for i := 0 to List.Count - 2 do
    s := s + List[i] + delim;
  s := s + List[List.Count - 1];
  if finalDelim then s := s + delim;
  Result := s;
end;

procedure ConvertStringToList(const s: string; List: TStrings; delim:Char = ';');
var
  len, start: Integer;
  t: string;

  function NextItem: string;
  var
    i: Integer;
    inquote: Boolean;
  begin
    i := start;
    inquote:=False;
    while (i <= Len) and ((inquote) or (s[i] <> delim)) do begin
          if s[i]='"' then inquote:=not inquote;
          Inc(i);
    end;
    if (start <> i) then Result := Copy(s, start, i - start)
    else Result := '';
    start := i + 1;
  end;
begin
  start := 1;
  len := Length(s);
  List.Clear;
  if (len = 0) then Exit;
  repeat
    t := NextItem;
    if (t <> '') then List.Add(t);
  until (start > Len);
end;

function ADOTypeToFieldType(const ADOType: DataTypeEnum; EnableBCD: Boolean = True): TFieldType;
begin
   case ADOType of
      adEmpty: Result := ftUnknown;
      adTinyInt, adSmallInt: Result := ftSmallint;
      adError, adInteger, adUnsignedInt: Result := ftInteger;
      adBigInt, adUnsignedBigInt: Result := ftLargeInt;
      adUnsignedTinyInt, adUnsignedSmallInt: Result := ftWord;
      adSingle, adDouble: Result := ftFloat;
      adCurrency: Result := ftBCD;
      adBoolean: Result := ftBoolean;
      adDBDate: Result := ftDate;
      adDBTime: Result := ftTime;
      adDate, adDBTimeStamp, adFileTime: Result := ftDateTime;
      adChar: Result := ftFixedChar;
      adVarChar, adVarWChar: Result := ftString;
      adBSTR, adWChar: Result := ftWideString;
      adLongVarChar, adLongVarWChar: Result := ftMemo;
      adLongVarBinary: Result := ftBlob;
      adBinary: Result := ftBytes;
      adVarBinary: Result := ftVarBytes;
      adChapter: Result := ftDataSet;
      adPropVariant, adVariant: Result := ftVariant;
      adIUnknown: Result := ftInterface;
      adIDispatch: Result := ftIDispatch;
      adGUID: Result := ftGUID;
      adDecimal, adNumeric, adVarNumeric:
         if EnableBCD then
            Result := ftBCD
         else
            Result := ftFloat;
      else
         Result := ftUnknown;
   end;
end;

function FieldTypeToADOType(const FieldType: TFieldType): DataTypeEnum;
begin
   case FieldType of
      ftUnknown: Result := adEmpty;
      ftString, ftWideString: Result := adVarWChar;
      ftSmallint: Result := adSmallint;
      ftInteger, ftAutoInc: Result := adInteger;
      ftWord: Result := adUnsignedSmallInt;
      ftBoolean: Result := adBoolean;
      ftFloat: Result := adDouble;
      ftCurrency, ftBCD: Result := adCurrency;
      ftDate: Result := adDate;
      //ftTime: Result := adDBTime;
      ftTime: Result:= adDate;
      //ftDateTime: Result := adDBTimeStamp;
      ftDateTime: Result := adDate;
      ftBytes: Result := adBinary;
      ftVarBytes: Result := adVarBinary;
      ftMemo: Result := adLongVarWChar;
      ftBlob, ftGraphic..ftTypedBinary: Result := adLongVarBinary;
      ftFixedChar: Result := adChar;
      ftLargeint: Result := adBigInt;
      ftVariant: Result := adVariant;
      ftInterface: Result := adIUnknown;
      ftIDispatch: Result := adIDispatch;
      ftGuid: Result := adGUID;
      else
         DatabaseError('No Matching fieldtype');
         Result := adEmpty;
   end;
end;

procedure IndexFieldsToStrings(ADOTab:TCustomADODataSet;List:TStringList);
var i,len,j:Integer;
    s,t:string;
begin
     
     if (not Assigned(ADOTab)) or (not Assigned(List)) then Exit;
     List.Clear;
     s := TADODataset(ADOTab).Indexfieldnames;
     len:=Length(s);
     i:=1;
     while (i<=len) do begin
           j:=i;
           while (j<=len) and (s[j]<>';') do inc(j);
           t:=Copy(s,i,j-i);
           list.Add(t);
           i:=j+1;
     end;
     for i:=0 to list.Count-1 do begin
         s:=trim(list[i]);
         len:=Length(s);
         if (len>5) and (comparetext(Copy(s,len-4,5),' DESC')=0) then begin
            list.objects[i]:=Pointer(-1);
            list[i]:=trim(Copy(s,1,len-5));
         end else list.objects[i]:=Pointer(0);
         if (len>4) and (comparetext(Copy(s,len-3,4),' ASC')=0) then
            list[i]:=trim(Copy(s,1,len-4));
     end;
end;

procedure StringsToIndexFields(ADOTab:TCustomADODataSet;List:TStringlist);
var s:string;
    i:Integer;
begin
     if (not Assigned(ADOTab)) or (not Assigned(List)) then Exit;
     s:='';
     for i:=0 to List.Count-1 do begin
         s:=s+list[i];
         if list.objects[i]=Pointer(-1) then s:=s+' DESC;'
         else s:=s+';';
     end;
     Delete(s,Length(s),1); //remove last ;
     TADOTable(ADOTab).IndexFieldNames := s;
end;


end.
