{***************************************************************
 *
 * Unit Name: AJB_ADOX
 * Purpose  : Provide components to create and manage Access databases
 * Author   : Andrew Baylis
 * Date     : 29/09/2001
 * History  : TADOXCatalog - creates and manages Access Databases
              TADOXTableTemplate - provides a TDataset descendant with no
                 functionality. Merely allows the defining of fields and
                 indexdefs so that a new database's tables can be defined from
                 these properties.

              29/9/01 - Added IADOXColumn interface to make accessing
                        ADO column properties easier.
              29/9/01 - constants and simple functions moved to ADO_Func unit
 *
 ****************************************************************}

unit AJB_ADOX;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ADOX_TLB, db, AJB_DS, ADODB, ADOInt, JRO_TLB, Variants, ADODB_TLB;

type
  // Interface to new column type that translates many ADOX properties
  // into more 'user-friendly' ones
  IADOXColumn = interface
    function GetADOColumn: _Column;
    function GetAllowZeroLength: Boolean;
    function GetAutoIncrementBy: Integer;
    function GetAutoIncrementSeed: Integer;
    function GetDataType: TFieldType;
    function GetDefaultExpr: string;
    function GetRequired: Boolean;
    function GetSize: Integer;
    function GetValidationRule: OLEVariant;
    function GetValidationText: string;
    function Get_Attributes: ColumnAttributesEnum;
    function Get_DefinedSize: Integer;
    function Get_NumericScale: Byte;
    function Get_ParentCatalog: _Catalog;
    function Get_Precision: Integer;
    function Get_RelatedColumn: WideString;
    function Get_SortOrder: SortOrderEnum;
    function Get_Type_: DataTypeEnum;
    procedure SetADOColumn(const Value: _Column);
    procedure SetAllowZeroLength(const Value: Boolean);
    procedure SetAutoIncrementBy(Value: Integer);
    procedure SetAutoIncrementSeed(Value: Integer);
    procedure SetDataType(Value: TFieldType);
    procedure SetDefaultExpr(const Value: string);
    procedure SetRequired(Value: Boolean);
    procedure SetSize(const Value: Integer);
    procedure SetValidationRule(const Value: OleVariant);
    procedure SetValidationText(const Value: string);
    procedure Set_Attributes(Value: ColumnAttributesEnum);
    procedure Set_DefinedSize(Value: Integer);
    procedure Set_Name(const Value: WideString);
    procedure Set_NumericScale(Value: Byte);
    procedure Set_ParentCatalog(const Value: _Catalog);
    procedure Set_Precision(Value: Integer);
    procedure Set_RelatedColumn(const Value: WideString);
    procedure Set_SortOrder(Value: SortOrderEnum);
    procedure Set_Type_(Value: DataTypeEnum);
    procedure Duplicate;
    function Get_Name: WideString;
    property ADOColumn: _Column read GetADOColumn write SetADOColumn;
    property AllowZeroLength: Boolean read GetAllowZeroLength write
      SetAllowZeroLength;
    property Attributes: ColumnAttributesEnum read Get_Attributes write Set_Attributes;
    property AutoIncrementBy: Integer read GetAutoIncrementBy write
      SetAutoIncrementBy;
    property AutoIncrementSeed: Integer read GetAutoIncrementSeed write
      SetAutoIncrementSeed;
    property DataType: TFieldType read GetDataType write SetDataType;
    property DefaultExpr: string read GetDefaultExpr write SetDefaultExpr;
    property DefinedSize: Integer read Get_DefinedSize write Set_DefinedSize;
    property Name: WideString read Get_Name write Set_Name;
    property NumericScale: Byte read Get_NumericScale write Set_NumericScale;
    property ParentCatalog: _Catalog read Get_ParentCatalog write Set_ParentCatalog;
    property Precision: Integer read Get_Precision write Set_Precision;
    property RelatedColumn: WideString read Get_RelatedColumn write Set_RelatedColumn;
    property Required: Boolean read GetRequired write SetRequired;
    property Size: Integer read GetSize write SetSize;
    property SortOrder: SortOrderEnum read Get_SortOrder write Set_SortOrder;
    property ValidationRule: OLEVariant read GetValidationRule write
      SetValidationRule;
    property ValidationText: string read GetValidationText write SetValidationText;
  end;

  TEngineType = (Auto, Access97, Access2000);
  TProvstring = string;

  TADOXCatalog = class(TComponent)
  private
    FActive: Boolean;
    FCatalog: _Catalog;
    FConnection: TADOConnection;
    FConnectionList: TStringList;
    FConnectionString: string;
    FDefaultConnection: TADOConnection;
    FEngineType: TEngineType;
    FTables: Tables;
    procedure AddTable(const TableName: string; Source: TDataset; fIndexDefs:
      TIndexDefs = nil);
    procedure CheckActive;
    function GetDatabase: TFilename;
    function GetEncrypt: Boolean;
    function GetMode: string;
    function GetPassword: string;
    function GetProvider: TProvstring;
    function GetTables: Tables;
    function GetUserID: string;
    procedure SetActive(Value: Boolean);
    procedure SetConnection(Value: TADOConnection);
    procedure SetConnectionString(const Value: string);
    procedure SetDatabase(Value: TFilename);
    procedure SetEncrypt(Value: Boolean);
    procedure SetEngineType(const Value: TEngineType);
    procedure SetMode(const Value: string);
    procedure SetPassword(const Value: string);
    procedure SetProvider(Value: TProvstring);
    procedure SetUserID(const Value: string);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AddField(const TableName: string; fldDef: TFieldDef): Boolean;
    procedure AddIndex(const TableName, IndexName, Fields: string; Options:
      TIndexOptions; const DescFields: string = '');
    function AddLinkedTable(DataSource, SourceTable: string; Newname: string = '';
      Provider: string = ''): Boolean;
    procedure AddQuery(const Name: string; SQL: string);
    procedure AddTableFrom(const Source: TDataset; Name: string = ''); overload;
    procedure AddTableWithIndex(const Source: TDataset; fIndexes: TIndexDefs;
      Name: string = '');
    procedure ChangeAllLinks(NewDatabase: string);
    function ChangeFieldDef(const TableName, Fieldname: string; newDef: TFieldDef):
      Boolean;
    function ChangeQueryText(const queryName: string; const Text: string): Boolean;
    procedure Close;
    procedure CompactAccessDB;
    procedure CreateAccessDB(Overwrite: Boolean = True);
    function DeleteField(const TableName, FieldName: string): Boolean;
    procedure DeleteQuery(const Name: string);
    procedure DeleteTable(const Name: string);
    function FindProcedure(const Name: string): Procedure_;
    function FindTable(const TableName: string): _Table;
    function FindView(const Name: string): View;
    function GetQueryText(const queryname: string): string;
    function MakeUniqueTableName(const trialname: string): string;
    procedure OpenAccessDB;
    property Active: Boolean read FActive write SetActive;
    property Catalog: _Catalog read FCatalog;
    property ConnectionString: string read FConnectionString write
      SetConnectionString;
    property Tables: Tables read GetTables;
  published
    property Connection: TADOConnection read FConnection write SetConnection;
    property Database: TFilename read GetDatabase write SetDatabase;
    property Encrypt: Boolean read GetEncrypt write SetEncrypt;
    property EngineType: TEngineType read FEngineType write SetEngineType;
    property Mode: string read GetMode write SetMode;
    property Password: string read GetPassword write SetPassword;
    property Provider: TProvstring read GetProvider write SetProvider;
    property UserID: string read GetUserID write SetUserID;
  end;

  TADOXTableTemplate = class(TAJBBaseDataset)
  private
    FIndexDefs: TIndexDefs;
    FTableName: string;
    procedure SetIndexDefs(const Value: TIndexDefs);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CreateTable;
  published
    property IndexDefs: TIndexDefs read FIndexDefs write SetIndexDefs;
    property TableName: string read FTableName write FTableName;
    property Active;
  end;

  TADOXColumn = class(TInterfacedObject, IADOXColumn, _Column)
  private
    FADOColumn: _Column;
    FCatalog: _Catalog;
    function GetADOColumn: _Column;
    function GetAllowZeroLength: Boolean;
    function GetAutoIncrementBy: Integer;
    function GetAutoIncrementSeed: Integer;
    function GetDataType: TFieldType;
    function GetDefaultExpr: string;
    function GetRequired: Boolean;
    function GetSize: Integer;
    function GetValidationRule: OLEVariant;
    function GetValidationText: string;
    function Get_Attributes: ColumnAttributesEnum;
    function Get_DefinedSize: Integer;
    function Get_NumericScale: Byte;
    function Get_ParentCatalog: _Catalog;
    function Get_Precision: Integer;
    function Get_RelatedColumn: WideString;
    function Get_SortOrder: SortOrderEnum;
    function Get_Type_: DataTypeEnum;
    procedure SetADOColumn(const Value: _Column);
    procedure SetAllowZeroLength(const Value: Boolean);
    procedure SetAutoIncrementBy(Value: Integer);
    procedure SetAutoIncrementSeed(Value: Integer);
    procedure SetDataType(Value: TFieldType);
    procedure SetDefaultExpr(const Value: string);
    procedure SetRequired(Value: Boolean);
    procedure SetSize(const Value: Integer);
    procedure SetValidationRule(const Value: OleVariant);
    procedure SetValidationText(const Value: string);
    procedure Set_Attributes(Value: ColumnAttributesEnum);
    procedure Set_DefinedSize(Value: Integer);
    procedure Set_Name(const Value: WideString);
    procedure Set_NumericScale(Value: Byte);
    procedure Set_ParentCatalog(const Value: _Catalog);
    procedure Set_Precision(Value: Integer);
    procedure Set_RelatedColumn(const Value: WideString);
    procedure Set_SortOrder(Value: SortOrderEnum);
    procedure Set_Type_(Value: DataTypeEnum);
    function Get_Name: WideString;
  protected
  public
    constructor Create(Catalog: _Catalog = nil);
    procedure Duplicate;
    property ADOColumn: _Column read GetADOColumn write SetADOColumn implements _Column;
    property AllowZeroLength: Boolean read GetAllowZeroLength write
      SetAllowZeroLength;
    property Attributes: ColumnAttributesEnum read Get_Attributes write Set_Attributes;
    property AutoIncrementBy: Integer read GetAutoIncrementBy write
      SetAutoIncrementBy;
    property AutoIncrementSeed: Integer read GetAutoIncrementSeed write
      SetAutoIncrementSeed;
    property DataType: TFieldType read GetDataType write SetDataType;
    property DefaultExpr: string read GetDefaultExpr write SetDefaultExpr;
    property DefinedSize: Integer read Get_DefinedSize write Set_DefinedSize;
    property Name: WideString read Get_Name write Set_Name;
    property NumericScale: Byte read Get_NumericScale write Set_NumericScale;
    property ParentCatalog: _Catalog read Get_ParentCatalog write Set_ParentCatalog;
    property Precision: Integer read Get_Precision write Set_Precision;
    property RelatedColumn: WideString read Get_RelatedColumn write Set_RelatedColumn;
    property Required: Boolean read GetRequired write SetRequired;
    property Size: Integer read GetSize write SetSize;
    property SortOrder: SortOrderEnum read Get_SortOrder write Set_SortOrder;
    property ValidationRule: OLEVariant read GetValidationRule write
      SetValidationRule;
    property ValidationText: string read GetValidationText write SetValidationText;
  end;

implementation

uses ADO_Func;

const
  cDataSource = 'Data Source';
  cMode = 'Mode';
  cPassword = 'Password';
  cProvider = 'Provider';
  cUserID = 'User ID';

  //************ AJB_ADOX:TADOXCatalog.AddTable *************************
  // Adds a table to an Access Database. Copies the definitions from Source
  // Uses the Fields if defined, otherwise Fielddefs.

procedure TADOXCatalog.AddTable(const TableName: string; Source: TDataset;
  fIndexDefs: TIndexDefs = nil);
var
  i, j: Integer;
  fldef: TFieldDef;
  fld: TField;
  idx: TIndexDef;
  list, descList: TStringList;

  mTab: _Table;
  mCol: IADOXColumn;
  mIndex: _Index;

begin
  if (not FActive) then
    raise EDatabaseError.Create('Need a target database');
  if (FindTable(TableName) <> nil) then
    raise EDatabaseError.Create('Table already exists');
  list := TStringList.Create;
  descList := TStringList.Create;
  try
    mTab := CoTable.Create;
    mTab.ParentCatalog := FCatalog;
    mTab.Name := TableName;
    if Source.FieldDefs.Count > Source.Fields.Count then
      begin
        for i := 0 to Source.FieldDefs.Count - 1 do
          begin
            fldef := Source.FieldDefs[i];
            mCol := TADOXColumn.Create(FCatalog); // sets ParentCatalog as well
            mCol.Name := fldef.Name;
            mCol.DataType := fldef.DataType;
            mCol.Size := fldef.Size;
            mCol.Required := fldef.Required;
            mCol.Precision := fldef.Precision;

            mTab.Columns.Append(mCol.ADOColumn, Unassigned, Unassigned);
          end;
      end
    else
      begin
        for i := 0 to Source.Fields.Count - 1 do
          if (Source.Fields[i].FieldKind = fkData) then
            begin
              fld := Source.Fields[i];
              mCol := TADOXColumn.Create(FCatalog); // sets ParentCatalog as well
              mCol.Name := fld.FieldName;
              mCol.DataType := fld.DataType;
              mCol.Size := fld.Size;
              mCol.Required := fld.Required;

              mCol.DefaultExpr := fld.DefaultExpression;
              mCol.ValidationRule := fld.CustomConstraint;
              mCol.ValidationText := fld.ConstraintErrorMessage;

              mTab.Columns.Append(mCol.ADOColumn, Unassigned, Unassigned);
            end;
      end;
    if (fIndexDefs <> nil) then
      for i := 0 to FIndexDefs.Count - 1 do
        begin
          idx := fIndexDefs[i];
          mIndex := ADOX_TLB.CoIndex.Create;
          mIndex.Name := idx.Name;
          ConvertStringToList(idx.Fields, list);
          ConvertStringToList(idx.DescFields, descList);

          for j := 0 to list.Count - 1 do
            begin
              mCol.ADOColumn := mTab.Columns[list[j]];
              mCol.Duplicate;
              if ((ixDescending in idx.Options) or (descList.IndexOf(list[j]) > -1)) then
                mCol.ADOColumn.SortOrder := adSortDescending;
              mIndex.Columns.Append(mCol.ADOColumn, Unassigned, Unassigned);
            end;
          if (ixPrimary in idx.Options) then
            mIndex.PrimaryKey := True
          else if (ixUnique in idx.Options) then
            mIndex.Unique := True;

          mTab.Indexes.Append(mIndex, Unassigned);
        end;
    FCatalog.Tables.Append(mTab);
  finally
    list.Free;
    descList.Free;
    FCatalog.Tables.Refresh;
  end;
end;

procedure TADOXCatalog.CheckActive;
begin
  if not Active then
    begin
      Active := true;
      if not Active then
        raise EDatabaseError.Create('Error in opening Database');
    end;
end;

function TADOXCatalog.GetDatabase: TFilename;
begin
  if FConnectionString <> '' then
    begin
      ConvertStringToList(FConnectionString, FConnectionList);
      Result := FConnectionList.Values[cDataSource];
    end
  else
    Result := '';
end;

function TADOXCatalog.GetEncrypt: Boolean;
begin
  if ConnectionString <> '' then
    begin
      ConvertStringToList(ConnectionString, FConnectionList);
      Result := (FConnectionList.Values[cJETEncrypt] = 'True');
    end
  else
    Result := False;
end;

function TADOXCatalog.GetMode: string;
begin
  if FConnectionString <> '' then
    begin
      ConvertStringToList(FConnectionString, FConnectionList);
      Result := FConnectionList.Values[cMode];
    end
  else
    Result := '';
end;

function TADOXCatalog.GetPassword: string;
begin
  if FConnectionString <> '' then
    begin
      ConvertStringToList(FConnectionString, FConnectionList);
      Result := FConnectionList.Values[cPassword];
    end
  else
    Result := '';
end;

function TADOXCatalog.GetProvider: TProvstring;
begin
  if FConnectionString <> '' then
    begin
      ConvertStringToList(FConnectionString, FConnectionList);
      Result := FConnectionList.Values[cProvider];
    end
  else
    Result := '';
end;

function TADOXCatalog.GetTables: Tables;
begin
  CheckActive;
  Result := FCatalog.Tables;
end;

function TADOXCatalog.GetUserID: string;
begin
  if FConnectionString <> '' then
    begin
      ConvertStringToList(FConnectionString, FConnectionList);
      Result := FConnectionList.Values[cUserID];
    end
  else
    Result := '';
end;

procedure TADOXCatalog.SetActive(Value: Boolean);
begin
  if (FActive <> Value) then
    begin
      if FActive then
        begin
          FCatalog.Set_ActiveConnection(Unassigned);
          FDefaultConnection.Connected := False;
          FCatalog := nil;
          FTables := nil;
        end
      else
        begin
          FCatalog := CoCatalog.Create;
          if Assigned(FConnection) and (FConnection.Connected) then
            FCatalog._Set_ActiveConnection(FConnection.ConnectionObject)
          else
            begin
              FDefaultConnection.Connected := False;
              FDefaultConnection.ConnectionString := ConnectionString;
              FDefaultConnection.Connected := True;
              FCatalog._Set_ActiveConnection(FDefaultConnection.ConnectionObject);
            end;
        end;
      FActive := Value;
    end;
end;

procedure TADOXCatalog.SetConnection(Value: TADOConnection);
begin
  FConnection := Value;
end;

procedure TADOXCatalog.SetConnectionString(const Value: string);
begin
  FConnectionString := Value;
end;

procedure TADOXCatalog.SetDatabase(Value: TFilename);
begin
  if (FConnectionString <> '') and (FConnectionList.IndexOfName(cDataSource) > -1) then
    FConnectionList.Values[cDataSource] := Value
  else
    FConnectionList.Add(cDataSource + '=' + Value);
  FConnectionString := ConvertListToString(FConnectionList);
end;

procedure TADOXCatalog.SetEncrypt(Value: Boolean);
var
  s: string;
  x: Integer;
begin
  if Value then
    s := 'True'
  else
    s := 'False';
  x := FConnectionList.IndexOfName(cJETEncrypt);
  if (ConnectionString <> '') and (x > -1) then
    begin
      if not Value then
        FConnectionList.Delete(x)
      else
        FConnectionList.Values[cJETEncrypt] := s;
    end
  else if Value then
    FConnectionList.Add(cJETEncrypt + '=' + s);
  ConnectionString := ConvertListToString(FConnectionList);
end;

procedure TADOXCatalog.SetEngineType(const Value: TEngineType);
var
  s: string;
  i: integer;
begin
  FEngineType := Value;
  i := FConnectionList.IndexOfName(Engine);
  if Value = Auto then
    begin
      if i > -1 then
        FConnectionList.Delete(i);
    end
  else
    begin
      if Value = Access97 then
        s := '4'
      else
        s := '5';
      if (FConnectionString <> '') and (i > -1) then
        FConnectionList.Values[Engine] := s
      else
        FConnectionList.Add(Engine + '=' + s);
    end;
  FConnectionString := ConvertListToString(FConnectionList);
end;

procedure TADOXCatalog.SetMode(const Value: string);
begin
  if (FConnectionString <> '') and (FConnectionList.IndexOfName(cMode) > -1) then
    FConnectionList.Values[cMode] := Value
  else
    FConnectionList.Add(cMode + '=' + Value);
  FConnectionString := ConvertListToString(FConnectionList);
end;

procedure TADOXCatalog.SetPassword(const Value: string);
begin
  if (FConnectionString <> '') and (FConnectionList.IndexOfName(cPassword) > -1) then
    FConnectionList.Values[cPassword] := Value
  else
    FConnectionList.Add(cPassword + '=' + Value);
  FConnectionString := ConvertListToString(FConnectionList);
end;

procedure TADOXCatalog.SetProvider(Value: TProvstring);
begin
  if (FConnectionString <> '') and (FConnectionList.IndexOfName(cProvider) > -1) then
    FConnectionList.Values[cProvider] := Value
  else
    FConnectionList.Add(cProvider + '=' + Value);
  FConnectionString := ConvertListToString(FConnectionList);
end;

procedure TADOXCatalog.SetUserID(const Value: string);
begin
  if (FConnectionString <> '') and (FConnectionList.IndexOfName(cUserID) > -1) then
    FConnectionList.Values[cUserID] := Value
  else
    FConnectionList.Add(cUserID + '=' + Value);
  FConnectionString := ConvertListToString(FConnectionList);
end;

procedure TADOXCatalog.Notification(AComponent: TComponent; Operation:
  TOperation);
begin
  if (Operation = opRemove) and (AComponent = FConnection) then
    FConnection := nil;
  inherited;
end;

constructor TADOXCatalog.Create(AOwner: TComponent);
begin
  inherited;
  FConnectionList := TStringList.Create;
  Provider := Jet4;
  Database := '';
  EngineType := Auto;
  FActive := False;
  FDefaultConnection := TADOConnection.Create(self);
  FdefaultConnection.LoginPrompt := False;
end;

destructor TADOXCatalog.Destroy;
begin
  Active := False;
  FConnectionList.Free;
  FDefaultConnection.Free;
  inherited;
end;

function TADOXCatalog.AddField(const TableName: string; fldDef: TFieldDef):
  Boolean;
var
  mTab: _Table;
  i: Integer;
  s: string;
  mCol: IADOXColumn;
begin
  CheckActive;
  Result := False;
  FCatalog.Tables.Refresh;
  i := 0;
  while (not Result) and (i < FCatalog.Tables.Count) do
    begin
      s := FCatalog.Tables.Item[i].Name;
      if (StrIComp(PChar(s), PChar(TableName)) = 0) then
        begin
          Result := True;
          mTab := FCatalog.Tables.Item[i];
        end;
      inc(i);
    end;

  if Result then
    begin
      mCol := TADOXColumn.Create(FCatalog); // sets ParentCatalog as well
      mCol.Name := fldDef.Name;
      mCol.DataType := fldDef.DataType;
      mCol.Size := fldDef.Size;
      mCol.Required := fldDef.Required;
      mCol.Precision := fldDef.Precision;
      mCol.AllowZeroLength := True;

      mTab.Columns.Append(mCol.ADOColumn, Unassigned, Unassigned);
      FCatalog.Tables.Refresh;
    end;
end;

procedure TADOXCatalog.AddIndex(const TableName, IndexName, Fields: string;
  Options: TIndexOptions; const DescFields: string = '');
var
  mCol: IADOXColumn;
  mTab: _Table;
  mIndex: _Index;
  j: Integer;
  list, descList: TStringList;
begin
  CheckActive;
  mTab := FindTable(TableName);
  if (mTab = nil) then
    Exit;

  list := TStringList.Create;
  descList := TStringList.Create;
  mCol := TADOXColumn.Create(FCatalog);

  try
    mIndex := ADOX_TLB.CoIndex.Create;
    mIndex.Name := IndexName;
    ConvertStringToList(Fields, list);
    ConvertStringToList(DescFields, descList);
    for j := 0 to list.Count - 1 do
      begin
        mCol.ADOColumn := mTab.Columns[list[j]];
        mCol.Duplicate;
        if ((ixDescending in Options) or (descList.IndexOf(list[j]) > -1)) then
          mCol.ADOColumn.SortOrder := adSortDescending;
        mIndex.Columns.Append(mCol.ADOColumn, Unassigned, Unassigned);
      end;
    if (ixPrimary in Options) then
      mIndex.PrimaryKey := True
    else if (ixUnique in Options) then
      mIndex.Unique := True;

    mTab.Indexes.Append(mIndex, Unassigned);
  finally
    list.Free;
    descList.Free;
  end;
end;

function TADOXCatalog.AddLinkedTable(DataSource, SourceTable: string; Newname:
  string = ''; Provider: string = ''): Boolean;
var
  mTab: _Table;

begin
  if (not FActive) then
    raise EDatabaseError.Create('Need a target database');
  if (DataSource = '') or (SourceTable = '') then
    raise EDatabaseError.Create('Insufficient information');
  mTab := CoTable.Create;
  mTab.ParentCatalog := FCatalog;

  if (Newname = '') then
    mTab.Name := MakeUniqueTableName(SourceTable)
  else
    mTab.Name := MakeUniqueTableName(NewName);

  // set up link details
  mTab.Properties[cJETLinkSource].Value := DataSource;
  if (Provider <> '') then
    mTab.Properties[cJETLinkProvider].Value := Provider;
  mTab.Properties[cJETRemoteTable].Value := SourceTable;
  mTab.Properties[cJETCreateLink].Value := True;

  // now add the table
  FCatalog.Tables.Append(mTab);
  Result := True; // success
  FCatalog.Tables.Refresh;
end;

procedure TADOXCatalog.AddQuery(const Name: string; SQL: string);
var
  mCmd: _Command;
begin
  if (not FActive) then
    raise EDatabaseError.Create('Need a target database');
  if (FindProcedure(Name) <> nil) or (FindView(Name) <> nil) then
    raise EDatabaseError.Create('Query already exists');
  mCmd := CoCommand.Create;
  mCmd.CommandText := SQL;
  mCmd.CommandType := adCmdText;
  mCmd.Name := Name;
  Catalog.Procedures.Append(Name, mCmd);
end;

procedure TADOXCatalog.AddTableFrom(const Source: TDataset; Name: string = '');
var
  sTableName: string;
  fIndexDefs: TIndexDefs;
begin
  fIndexDefs := nil;
  sTableName := Name;
  if sTableName = '' then
    sTableName := Source.Name;
  if (Source is TADOTable) then
    begin
      fIndexDefs := TADOTable(Source).IndexDefs;
      if (TADOTable(Source).TableName <> '') then
        sTableName := TADOTable(Source).TableName;
    end
  else if (Source is TADOXTableTemplate) then
    begin
      fIndexDefs := TADOXTableTemplate(Source).IndexDefs;
      if (TADOXTableTemplate(Source).TableName <> '') then
        sTableName := TADOXTableTemplate(Source).TableName;
    end;

  AddTable(sTableName, Source, fIndexDefs);
end;

procedure TADOXCatalog.AddTableWithIndex(const Source: TDataset; fIndexes:
  TIndexDefs; Name: string = '');
var
  sTableName: string;
begin
  sTableName := Name;
  if sTableName = '' then
    sTableName := Source.Name;
  AddTable(sTableName, Source, fIndexes);
end;

procedure TADOXCatalog.ChangeAllLinks(NewDatabase: string);
var
  i, j: Integer;
begin
  CheckActive;
  j := 0;
  for i := 0 to FCatalog.Tables.Count - 1 do
    if FCatalog.Tables[i].Type_ = 'LINK' then
      if not AnsiSameText(FCatalog.Tables[i].Properties[cJETLinkSource].Value, NewDatabase) then
        begin
          FCatalog.Tables[i].Properties[cJETLinkSource].Value := NewDatabase;
          Inc(j);
        end;
  //only refresh if changes have been made
  if (j > 0) then
    FCatalog.Tables.Refresh;
end;

function TADOXCatalog.ChangeFieldDef(const TableName, Fieldname: string;
  newDef: TFieldDef): Boolean;
var
  mTab: _Table;
  i: Integer;
  s: string;
  mCol: IADOXColumn;
begin
  CheckActive;
  Result := False;
  FCatalog.Tables.Refresh;
  i := 0;
  while (not Result) and (i < FCatalog.Tables.Count) do
    begin
      s := FCatalog.Tables.Item[i].Name;
      if (StrIComp(PChar(s), PChar(TableName)) = 0) then
        begin
          Result := True;
          mTab := FCatalog.Tables.Item[i];
        end;
      inc(i);
    end;

  if Result then
    begin
      mCol := TADOXColumn.Create(FCatalog); // sets ParentCatalog as well
      Result := False;
      while (not Result) and (i < mTab.Columns.Count) do
        begin
          s := mTab.Columns.Item[i].Name;
          if (StrIComp(PChar(s), PChar(FieldName)) = 0) then
            begin
              Result := True;
              mCol.ADOColumn := mTab.Columns[i];
            end;
          inc(i);
        end;
      if Result then
        begin
          mCol.Name := newDef.Name;
          mCol.DataType := newDef.DataType;
          mCol.Size := newDef.Size;
          mCol.Required := newDef.Required;
          mCol.Precision := newDef.Precision;
          mCol.AllowZeroLength := True;
        end;
      FCatalog.Tables.Refresh;
    end;
end;

function TADOXCatalog.ChangeQueryText(const queryName: string; const
  Text: string): Boolean;
var
  mCmd: _Command;
  mView: View;
  mProc: Procedure_;
begin
  if (not FActive) then
    raise EDatabaseError.Create('Need a target database');
  mCmd := CoCommand.Create;
  mCmd.CommandText := Text;
  mCmd.CommandType := adCmdText;
  mCmd.Name := queryName;

  mView := FindView(queryName);
  if (mView <> nil) then
    begin
      mView._Set_Command(mCmd);
      Result := True;
    end
  else
    begin
      mProc := FindProcedure(queryName);
      if (mProc <> nil) then
        begin
          mProc._Set_Command(mCmd);
          Result := True;
        end
      else
        raise EDatabaseError.Create('Query doesn''t exist');
    end;
end;

procedure TADOXCatalog.Close;
begin
  Active := False;
end;

procedure TADOXCatalog.CompactAccessDB;
//note any connections to the database will be closed
var
  jet: IJetEngine;
  Destination, Source, OldSource: string;
  currentdb, tempdb: string;

begin
  Active := False;
  if Assigned(FCatalog) then
    FCatalog := nil; //double check!!
  OldSource := ConnectionString;

  jet := CoJetEngine.Create;
  try
    if Assigned(FConnection) then
      begin
        Source := FConnection.ConnectionString;
        FConnection.Connected := False; //ensure no active connections to source
      end
    else
      begin
        if not FileExists(Database) then
          raise EFCreateError.CreateHelp(Database + ' doesn''t exist', 0);
        Source := ConnectionString;
      end;
    ConnectionString := Source;
    currentdb := Database;
    tempdb := ChangeFileExt(currentdb, '.$mb');

    Database := tempdb;
    Destination := ConnectionString; //generated by setting database
    Database := currentdb;
    try
      jet.CompactDatabase(Source, Destination);
      if DeleteFile(currentdb) then
        RenameFile(tempdb, currentdb);

    except
      MessageDlg('Error in compacting database.' +
        #13 + #10 + 'No changes made.', mtError, [mbYes], 0);
    end;
  finally
    jet := nil;
    ConnectionString := OldSource; //restore to original state
    if FileExists(tempdb) then
      DeleteFile(tempdb); //remove tempdb if an error occurred
  end;
end;

procedure TADOXCatalog.CreateAccessDB(Overwrite: Boolean = True);
begin
  if FileExists(Database) then
    begin
      if Overwrite then
        begin
          if not DeleteFile(Database) then
            raise EFCreateError.CreateHelp(Database + ' cannot be overwritten!', 0);
        end
      else
        raise EFCreateError.CreateHelp(Database + ' already exists', 0);
    end;
  Active := True;
  Catalog.Create(ConnectionString);
end;

function TADOXCatalog.DeleteField(const TableName, FieldName: string): Boolean;
var
  mTab: _Table;
  i: Integer;
  s: string;

begin
  CheckActive;
  Result := False;
  FCatalog.Tables.Refresh;
  i := 0;
  while (not Result) and (i < FCatalog.Tables.Count) do
    begin
      s := FCatalog.Tables.Item[i].Name;
      if (StrIComp(PChar(s), PChar(TableName)) = 0) then
        begin
          Result := True;
          mTab := FCatalog.Tables.Item[i];
        end;
      inc(i);
    end;

  if Result then
    begin
      Result := False;
      i := 0;
      while (not Result) and (i < mTab.Columns.Count) do
        begin
          s := mTab.Columns.Item[i].Name;
          if (StrIComp(PChar(s), PChar(FieldName)) = 0) then
            begin
              Result := True;
              mTab.Columns.Delete(i);
            end;
          inc(i);
        end;
      FCatalog.Tables.Refresh;
    end;
end;

procedure TADOXCatalog.DeleteQuery(const Name: string);
begin
  if (FindProcedure(Name) <> nil) then
    Catalog.Procedures.Delete(Name)
  else if (FindView(Name) <> nil) then
    Catalog.Views.Delete(Name);
end;

procedure TADOXCatalog.DeleteTable(const Name: string);
begin
  if (FindTable(Name) <> nil) then
    Catalog.Tables.Delete(Name);
end;

function TADOXCatalog.FindProcedure(const Name: string): Procedure_;
var
  i, qrycount: Integer;
begin
  Result := nil;
  if FActive then
    begin
      qrycount := FCatalog.Procedures.Count;
      i := 0;
      while (i < qrycount) and (CompareText(FCatalog.Procedures[i].Name, Name) <> 0) do
        Inc(i);
      if (i < qrycount) then
        Result := FCatalog.Procedures[i];
    end;
end;

function TADOXCatalog.FindTable(const TableName: string): _Table;
var
  i, tabcount: Integer;
begin
  Result := nil;
  if FActive then
    begin
      tabcount := FCatalog.Tables.Count;
      i := 0;
      while (i < tabcount) and
        (CompareText(FCatalog.Tables[i].Name, TableName) <> 0) do
        Inc(i);
      if (i < tabcount) then
        Result := FCatalog.Tables[i];
    end;
end;

function TADOXCatalog.FindView(const Name: string): View;
var
  i, qrycount: Integer;
begin
  Result := nil;
  if FActive then
    begin
      qrycount := FCatalog.Views.Count;
      i := 0;
      while (i < qrycount) and (CompareText(FCatalog.Views[i].Name, Name) <> 0) do
        Inc(i);
      if (i < qrycount) then
        Result := FCatalog.Views[i];
    end;
end;

function TADOXCatalog.GetQueryText(const queryname: string): string;
var
  mCmd: OleVariant;
  mView: View;
  mProc: Procedure_;
begin
  Result := '';
  if (not FActive) then
    raise EDatabaseError.Create('Need a target database');

  mView := FindView(queryName);
  if (mView <> nil) then
    begin
      mCmd := mView.Get_Command;
      if mCmd.CommandType = adCmdText then
        Result := mCmd.CommandText;
    end
  else
    begin
      mProc := FindProcedure(queryName);
      if (mProc <> nil) then
        begin
          mCmd := mProc.Get_Command;
          if mCmd.CommandType = adCmdText then
            Result := mCmd.CommandText;
        end
      else
        raise EDatabaseError.Create('Query doesn''t exist');
    end;
end;

function TADOXCatalog.MakeUniqueTableName(const trialname: string): string;
var
  i: Integer;
begin
  Result := trialname;
  i := 1;
  while (FindTable(Result) <> nil) do
    begin
      Result := trialname + IntToStr(i);
      Inc(i);
    end;
end;

procedure TADOXCatalog.OpenAccessDB;
begin
  if not (Assigned(FConnection) and (FConnection.Connected)) then
    if not FileExists(Database) then
      raise EFCreateError.CreateHelp(Database + ' doesn''t exist', 0);
  Active := True;
  FTables := nil;
end;

procedure TADOXTableTemplate.SetIndexDefs(const Value: TIndexDefs);
begin
  FIndexDefs.Assign(Value);
end;

constructor TADOXTableTemplate.Create(AOwner: TComponent);
begin
  inherited;
  FIndexDefs := TIndexDefs.Create(Self);
  TableName := '';
end;

destructor TADOXTableTemplate.Destroy;
begin
  FIndexDefs.Free;
  inherited;
end;

procedure TADOXTableTemplate.CreateTable;
var
  i: Integer;
begin
  if Active then Active := False;

  // If no fielddefs existing, use the previously defined fields.
  if FieldDefs.Count = 0 then
    for i := 0 to FieldCount - 1 do
      with Fields[i] do
        if FieldKind = fkData then
          FieldDefs.Add(FieldName, DataType, Size, Required);

  // Remove previously defined fields.
  DestroyFields;

  // Create fields.
  CreateFields;
  Active := True;
end;

function TADOXColumn.GetADOColumn: _Column;
begin
  Result := FADOColumn;
end;

function TADOXColumn.GetAllowZeroLength: Boolean;
begin
  if (DataType in [ftString, ftWideString]) and (FADOColumn.Properties.Count > 0) then
    Result := FADOColumn.Properties[cJETAllowZeroLength].Value
  else
    Result := False;
end;

function TADOXColumn.GetAutoIncrementBy: Integer;
begin
  if (FADOColumn.Properties.Count > 0) then
    Result := FADOColumn.Properties[cAutoIncrement].Value
  else
    Result := 1;
end;

function TADOXColumn.GetAutoIncrementSeed: Integer;
begin
  if (FADOColumn.Properties.Count > 0) then
    Result := FADOColumn.Properties[cAutoSeed].Value
  else
    Result := 1;
end;

function TADOXColumn.GetDataType: TFieldType;
begin
  Result := ADOTypeToFieldType(FADOColumn.Type_);
  if (Result = ftInteger) and (FADOColumn.Properties.Count > 0) and (VarAsType(FADOColumn.Properties[cAutoInc].Value, varBoolean) = True) then
    Result := ftAutoInc;
end;

function TADOXColumn.GetDefaultExpr: string;
begin
  if FADOColumn.Properties.Count > 0 then
    Result := FADOColumn.Properties[cDefault].Value
  else
    Result := '';
end;

function TADOXColumn.GetRequired: Boolean;
begin
  if (FADOColumn.Properties.Count > 0) then
    Result := not FADOColumn.Properties[cNullable].Value
  else
    Result := False;
end;

function TADOXColumn.GetSize: Integer;
begin
  Result := FADOColumn.DefinedSize;
end;

function TADOXColumn.GetValidationRule: OLEVariant;
begin
  if (FADOColumn.Properties.Count > 0) then
    Result := FADOColumn.Properties[cJETValidateRule].Value
  else
    Result := EmptyParam;
end;

function TADOXColumn.GetValidationText: string;
begin
  if FADOColumn.Properties.Count > 0 then
    Result := FADOColumn.Properties[cJETValidateText].Value
  else
    Result := '';
end;

function TADOXColumn.Get_Attributes: ColumnAttributesEnum;
begin
  Result := FADOColumn.Attributes;
end;

function TADOXColumn.Get_DefinedSize: Integer;
begin
  Result := FADOColumn.DefinedSize;
end;

function TADOXColumn.Get_NumericScale: Byte;
begin
  Result := FADOColumn.NumericScale;
end;

function TADOXColumn.Get_ParentCatalog: _Catalog;
begin
  Result := FADOColumn.Get_ParentCatalog;
end;

function TADOXColumn.Get_Precision: Integer;
begin
  Result := FADOColumn.Get_Precision;
end;

function TADOXColumn.Get_RelatedColumn: WideString;
begin
  Result := FADOColumn.Get_RelatedColumn;
end;

function TADOXColumn.Get_SortOrder: SortOrderEnum;
begin
  Result := FADOColumn.Get_SortOrder;
end;

function TADOXColumn.Get_Type_: DataTypeEnum;
begin
  Result := FADOColumn.Type_;
end;

procedure TADOXColumn.SetADOColumn(const Value: _Column);
begin
  FADOColumn := Value;
end;

procedure TADOXColumn.SetAllowZeroLength(const Value: Boolean);
begin
  if (DataType in [ftString, ftWideString]) and (FADOColumn.Properties.Count > 0) then
    FADOColumn.Properties[cJETAllowZeroLength].Value := Value;
end;

procedure TADOXColumn.SetAutoIncrementBy(Value: Integer);
begin
  if (FADOColumn.Properties.Count > 0) then
    FADOColumn.Properties[cAutoIncrement].Value := Value;
end;

procedure TADOXColumn.SetAutoIncrementSeed(Value: Integer);
begin
  if (FADOColumn.Properties.Count > 0) then
    FADOColumn.Properties[cAutoSeed].Value := Value;
end;

procedure TADOXColumn.SetDataType(Value: TFieldType);
begin
  FADOColumn.Type_ := FieldTypeToADOType(Value);
  case Value of
    ftAutoInc:
      if (FADOColumn.Properties.Count > 0) then
        FADOColumn.Properties[cAutoInc].Value := True;
    ftString, ftWideString: Size := 50;
    ftBoolean: FADOColumn.Attributes := adColFixed;
  end;
end;

procedure TADOXColumn.SetDefaultExpr(const Value: string);
begin
  if (DataType in sValidate) and (FADOColumn.Properties.Count > 0) then
    FADOColumn.Properties[cDefault].Value := Value;
end;

procedure TADOXColumn.SetRequired(Value: Boolean);
begin
  if (Datatype in sNullable) and (FADOColumn.Properties.Count > 0) then
    FADOColumn.Properties[cNullable].Value := not Value;
end;

procedure TADOXColumn.SetSize(const Value: Integer);
begin
  FADOColumn.DefinedSize := Value;
end;

procedure TADOXColumn.SetValidationRule(const Value: OLEVariant);
begin
  if (DataType in sValidate) and (FADOColumn.Properties.Count > 0) then
    FADOColumn.Properties[cJETValidateRule].Value := Value;
end;

procedure TADOXColumn.SetValidationText(const Value: string);
begin
  if (DataType in sValidate) and (FADOColumn.Properties.Count > 0) then
    FADOColumn.Properties[cJETValidateText].Value := Value;
end;

procedure TADOXColumn.Set_Attributes(Value: ColumnAttributesEnum);
begin
  FADOColumn.Attributes := Value;
end;

procedure TADOXColumn.Set_DefinedSize(Value: Integer);
begin
  FADOColumn.DefinedSize := Value;
end;

procedure TADOXColumn.Set_Name(const Value: WideString);
begin
  FADOColumn.Name := Value;
end;

procedure TADOXColumn.Set_NumericScale(Value: Byte);
begin
  FADOColumn.NumericScale := Value;
end;

procedure TADOXColumn.Set_ParentCatalog(const Value: _Catalog);
begin
  FADOColumn.ParentCatalog := Value;
  FCatalog := Value;
end;

procedure TADOXColumn.Set_Precision(Value: Integer);
begin
  FADOColumn.Precision := Value;
end;

procedure TADOXColumn.Set_RelatedColumn(const Value: WideString);
begin
  FADOColumn.RelatedColumn := Value;
end;

procedure TADOXColumn.Set_SortOrder(Value: SortOrderEnum);
begin
  FADOColumn.SortOrder := Value;
end;

procedure TADOXColumn.Set_Type_(Value: DataTypeEnum);
begin
  FADOColumn.Type_ := Value;
end;

constructor TADOXColumn.Create(Catalog: _Catalog = nil);
begin
  inherited Create;
  FADOColumn := CoColumn.Create;
  // ParentCatalog must be set to access the Properties collection
  if Catalog <> nil then
    FADOColumn.ParentCatalog := Catalog;
  FCatalog := Catalog;
end;

//************ AJB_ADOX:TADOXColumn.Duplicate *************************
// This procedure generates a new column with the same properties as
// the current one. It is useful when appending to Index.Columns or
// Keys.Columns as a column already in a collection cannot be appended

procedure TADOXColumn.Duplicate;
var
  tempCol: _Column;
  i: Integer;
begin
  tempCol := CoColumn.Create;
  // copy all properties to new one
  with FADOColumn do
    begin
      // For some reason ADOX returns nil when querying ParentCatalog
      // hence we need to use a previously stored version
      tempCol.ParentCatalog := FCatalog;
      tempCol.Name := Name;
      tempCol.DefinedSize := DefinedSize;
      tempCol.Attributes := Attributes;
      tempCol.NumericScale := NumericScale;
      tempCol.Precision := Precision;
      tempCol.Type_ := Type_;
      // note: don't copy SortOrder or RelatedColumn as these only apply
      // to columns in indexes or keys
      for i := 0 to tempCol.Properties.Count - 1 do
        tempCol.Properties[i].Value := Properties[i].Value;
    end;
  FADOColumn := tempCol; // replace with copy
end;

function TADOXColumn.Get_Name: WideString;
begin
  Result := FADOColumn.Name;
end;

end.

